from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

import xlwings as xw

this_dir = Path(__file__).resolve().parent

nrows, ncols = 5, 4
np_data = np.arange(nrows * ncols).reshape(nrows, ncols).astype(float)
list_data = np_data.tolist()
df = pd.DataFrame(
    data=np_data,
    columns=[f"c{i}" for i in range(ncols)],
    index=[f"c{i}" for i in range(nrows)],
)


@pytest.fixture(scope="module")
def app():
    with xw.App(visible=False) as app:
        yield app


def test_read_write_df(app):
    sheet = app.books[0].sheets[0]

    sheet["A1"].value = df
    assert_frame_equal(df, sheet["A1"].expand().options(pd.DataFrame).value)
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=2).value = df
    assert_frame_equal(
        df, sheet["A1"].expand().options(pd.DataFrame, chunksize=2).value
    )
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=nrows - 1).value = df
    assert_frame_equal(
        df, sheet["A1"].expand().options(pd.DataFrame, chunksize=nrows - 1).value
    )
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=nrows + 1).value = df
    assert_frame_equal(
        df, sheet["A1"].expand().options(pd.DataFrame, chunksize=nrows + 2).value
    )
    sheet["A1"].expand().clear()


def test_read_write_np(app):
    sheet = app.books[0].sheets[0]

    sheet["A1"].value = np_data
    assert_array_equal(np_data, sheet["A1"].expand().options(np.array).value)
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=2).value = np_data
    assert_array_equal(
        np_data, sheet["A1"].expand().options(np.array, chunksize=2).value
    )
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=nrows - 1).value = np_data
    assert_array_equal(
        np_data, sheet["A1"].expand().options(np.array, chunksize=nrows - 1).value
    )
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=nrows + 1).value = np_data
    assert_array_equal(
        np_data, sheet["A1"].expand().options(np.array, chunksize=nrows + 2).value
    )
    sheet["A1"].expand().clear()


def test_read_write_list(app):
    sheet = app.books[0].sheets[0]

    sheet["A1"].value = list_data
    for i in range(len(list_data)):
        assert list_data[i] == sheet["A1"].expand().value[i]
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=2).value = list_data
    for i in range(len(list_data)):
        assert list_data[i] == sheet["A1"].expand().options(chunksize=2).value[i]
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=nrows - 1).value = list_data
    for i in range(len(list_data)):
        assert (
            list_data[i] == sheet["A1"].expand().options(chunksize=nrows - 1).value[i]
        )
    sheet["A1"].expand().clear()

    sheet["A1"].options(chunksize=nrows + 1).value = list_data
    for i in range(len(list_data)):
        assert (
            list_data[i] == sheet["A1"].expand().options(chunksize=nrows + 2).value[i]
        )
    sheet["A1"].expand().clear()



# Characters are currently not properly supported
# on macOS due to an Excel/AppleScript bug
from pathlib import Path

import pytest

import xlwings as xw
from xlwings.pro import Markdown, MarkdownStyle

this_dir = Path(__file__).resolve().parent


text1 = """\
# Title

Text **bold** and *italic*

* a first bullet
* a second bullet

# Another title

this has a line break
new line
"""


@pytest.fixture(scope="module")
def app():
    with xw.App(visible=False) as app:
        app.books.open(this_dir / "test book.xlsx")
        yield app


# Range
def test_markdown_cell_defaults_formatting(app):
    cell = app.books["test book.xlsx"].sheets[0]["A1"]
    cell.clear()
    cell.value = Markdown(text1)
    assert cell.font.name == "Calibri"
    assert cell.font.size == 11
    assert cell.characters[:5].font.bold is True
    assert cell.characters[7:11].font.bold is False
    assert cell.characters[7:11].font.italic is False
    assert cell.characters[21:27].font.italic is True
    assert cell.characters[65:78].font.bold is True


def test_markdown_cell_defaults_value(app):
    cell = app.books["test book.xlsx"].sheets[0]["A1"]
    cell.clear()
    cell.value = Markdown(text1)
    assert (
        cell.value
        == "Title\n\nText bold and italic\n\n• a first bullet\n• a second bullet\n\n"
        "Another title\n\nthis has a line break\nnew line"
    )


def test_markdown_cell_h1(app):
    cell = app.books["test book.xlsx"].sheets[0]["A1"]
    cell.clear()
    style = MarkdownStyle()
    style.h1.blank_lines_after = 2
    style.h1.font.color = (255, 0, 0)
    style.h1.font.bold = False
    style.h1.font.size = 20
    style.h1.font.italic = True
    style.h1.font.name = "Arial"
    cell.value = Markdown(text1, style)

    for selection in [slice(0, 5), slice(68, 81)]:
        assert cell.characters[selection].font.color == (255, 0, 0)
        assert cell.characters[selection].font.bold is False
        assert cell.characters[selection].font.size == 20
        assert cell.characters[selection].font.italic is True
        assert cell.characters[selection].font.name == "Arial"


def test_markdown_cell_strong(app):
    cell = app.books["test book.xlsx"].sheets[0]["A1"]
    cell.clear()
    style = MarkdownStyle()
    style.strong.color = (255, 0, 0)
    style.strong.bold = False
    style.strong.size = 20
    style.strong.italic = True
    style.strong.name = "Arial"
    cell.value = Markdown(text1, style)

    assert cell.characters[12:16].font.color == (255, 0, 0)
    assert cell.characters[12:16].font.bold is False
    assert cell.characters[12:16].font.size == 20
    assert cell.characters[12:16].font.italic is True
    assert cell.characters[12:16].font.name == "Arial"


def test_markdown_cell_emphasis(app):
    cell = app.books["test book.xlsx"].sheets[0]["A1"]
    cell.clear()
    style = MarkdownStyle()
    style.emphasis.color = (255, 0, 0)
    style.emphasis.bold = False
    style.emphasis.size = 20
    style.emphasis.italic = True
    style.emphasis.name = "Arial"
    cell.value = Markdown(text1, style)

    assert cell.characters[21:27].font.color == (255, 0, 0)
    assert cell.characters[21:27].font.bold is False
    assert cell.characters[21:27].font.size == 20
    assert cell.characters[21:27].font.italic is True
    assert cell.characters[21:27].font.name == "Arial"


def test_markdown_cell_unordered_list(app):
    cell = app.books["test book.xlsx"].sheets[0]["A1"]
    cell.clear()
    style = MarkdownStyle()
    style.unordered_list.bullet_character = "-"
    style.unordered_list.blank_lines_after = 1
    cell.value = Markdown(text1, style)

    assert cell.characters[29].text == "-"
    assert cell.characters[65:72].text == "Another"

    style.unordered_list.blank_lines_after = 2
    cell.clear()
    cell.value = Markdown(text1, style)
    assert cell.characters[66:73].text == "Another"


# Shape
def test_markdown_shape_defaults_formatting(app):
    shape = app.books["test book.xlsx"].sheets["shape"].shapes[0]
    shape.text = ""
    shape.text = Markdown(text1)
    assert shape.font.name == "Calibri"
    assert shape.font.size == 11
    assert shape.characters[:5].font.bold is True
    assert shape.characters[7:11].font.bold is False
    assert shape.characters[7:11].font.italic is False
    assert shape.characters[21:27].font.italic is True
    assert shape.characters[65:78].font.bold is True


def test_markdown_shape_defaults_value(app):
    shape = app.books["test book.xlsx"].sheets["shape"].shapes[0]
    shape.text = ""
    shape.text = Markdown(text1)
    assert (
        shape.text
        == "Title\n\nText bold and italic\n\n• a first bullet\n• a second bullet\n\n"
        "Another title\n\nthis has a line break\nnew line"
    )


def test_markdown_shape_h1(app):
    shape = app.books["test book.xlsx"].sheets["shape"].shapes[0]
    shape.text = ""
    style = MarkdownStyle()
    style.h1.blank_lines_after = 2
    style.h1.font.color = (255, 0, 0)
    style.h1.font.bold = False
    style.h1.font.size = 20
    style.h1.font.italic = True
    style.h1.font.name = "Arial"
    shape.text = Markdown(text1, style)

    for selection in [slice(0, 5), slice(68, 81)]:
        assert shape.characters[selection].font.color == (255, 0, 0)
        assert shape.characters[selection].font.bold is False
        assert shape.characters[selection].font.size == 20
        assert shape.characters[selection].font.italic is True
        assert shape.characters[selection].font.name == "Arial"


def test_markdown_shape_strong(app):
    shape = app.books["test book.xlsx"].sheets["shape"].shapes[0]
    shape.text = ""
    style = MarkdownStyle()
    style.strong.color = (255, 0, 0)
    style.strong.bold = False
    style.strong.size = 20
    style.strong.italic = True
    style.strong.name = "Arial"
    shape.text = Markdown(text1, style)

    assert shape.characters[12:16].font.color == (255, 0, 0)
    assert shape.characters[12:16].font.bold is False
    assert shape.characters[12:16].font.size == 20
    assert shape.characters[12:16].font.italic is True
    assert shape.characters[12:16].font.name == "Arial"


def test_markdown_shape_emphasis(app):
    shape = app.books["test book.xlsx"].sheets["shape"].shapes[0]
    shape.text = ""
    style = MarkdownStyle()
    style.emphasis.color = (255, 0, 0)
    style.emphasis.bold = False
    style.emphasis.size = 20
    style.emphasis.italic = True
    style.emphasis.name = "Arial"
    shape.text = Markdown(text1, style)

    assert shape.characters[21:27].font.color == (255, 0, 0)
    assert shape.characters[21:27].font.bold is False
    assert shape.characters[21:27].font.size == 20
    assert shape.characters[21:27].font.italic is True
    assert shape.characters[21:27].font.name == "Arial"


def test_markdown_shape_unordered_list(app):
    shape = app.books["test book.xlsx"].sheets["shape"].shapes[0]
    shape.text = ""
    style = MarkdownStyle()
    style.unordered_list.bullet_character = "-"
    style.unordered_list.blank_lines_after = 1
    shape.text = Markdown(text1, style)

    assert shape.characters[29].text == "-"
    assert shape.characters[65:72].text == "Another"

    style.unordered_list.blank_lines_after = 2
    shape.text = ""
    shape.text = Markdown(text1, style)
    assert shape.characters[66:73].text == "Another"



import os
import sys
import unittest

import xlwings as xw

from .common import TestBase, this_dir

try:
    import numpy as np
except ImportError:
    np = None

try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
except ImportError:
    mpl = None

try:
    import PIL
except ImportError:
    PIL = None

if sys.version_info[0] >= 3 and sys.version_info[1] >= 6:
    import pathlib
else:
    pathlib = None

try:
    import plotly.graph_objects as plotly_go
except ImportError:
    plotly_go = None


class TestShape(TestBase):
    def test_name(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        self.wb1.sheets[0].pictures.add(filename, name="pic1")

        sh = self.wb1.sheets[0].shapes[0]
        self.assertEqual(sh.name, "pic1")
        sh.name = "yoyoyo"
        self.assertEqual(sh.name, "yoyoyo")

    @unittest.skipIf(pathlib is None, "pathlib unavailable")
    def test_name_pathlib(self):
        filename = pathlib.Path(this_dir) / "sample_picture.png"
        self.wb1.sheets[0].pictures.add(filename, name="pic1")

        sh = self.wb1.sheets[0].shapes[0]
        self.assertEqual(sh.name, "pic1")
        sh.name = "yoyoyo"
        self.assertEqual(sh.name, "yoyoyo")

    def test_coordinates(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        self.wb1.sheets[0].pictures.add(
            filename, name="pic1", left=0, top=0, width=200, height=100
        )

        sh = self.wb1.sheets[0].shapes[0]
        for a, init, neu in (
            ("left", 0, 50),
            ("top", 0, 50),
            ("width", 200, 150),
            ("height", 100, 160),
        ):
            self.assertEqual(getattr(sh, a), init)
            setattr(sh, a, neu)
            self.assertEqual(getattr(sh, a), neu)

    def test_picture_object(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        self.wb1.sheets[0].pictures.add(filename, name="pic1")

        self.assertEqual(
            self.wb1.sheets[0].shapes[0], self.wb1.sheets[0].shapes["pic1"]
        )

    def test_delete(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertTrue("pic1" in self.wb1.sheets[0].shapes)
        self.wb1.sheets[0].shapes[0].delete()
        self.assertFalse("pic1" in self.wb1.sheets[0].shapes)

    def test_type(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(self.wb1.sheets[0].shapes[0].type, "picture")

    def test_scale_width(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        w, h = int(pic.width), int(pic.height)
        self.wb1.sheets[0].shapes["pic1"].scale_width(factor=2)
        self.assertEqual(int(pic.width), w * 2)
        self.assertEqual(int(pic.height), h * 2)

    def test_scale_height(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        w, h = int(pic.width), int(pic.height)
        self.wb1.sheets[0].shapes["pic1"].scale_height(factor=2)
        self.assertEqual(int(pic.width), w * 2)
        self.assertEqual(int(pic.height), h * 2)


class TestPicture(TestBase):
    def test_two_books(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        pic2 = self.wb2.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(pic1.name, "pic1")
        self.assertEqual(pic2.name, "pic1")

    def test_name(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(pic.name, "pic1")

        pic.name = "pic_new"
        self.assertEqual(pic.name, "pic_new")

    def test_left(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(pic.left, 0)

        pic.left = 20
        self.assertEqual(pic.left, 20)

    def test_top(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(pic.left, 0)

        pic.top = 20
        self.assertEqual(pic.top, 20)

    def test_width(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(int(pic.width), 30)
        pic.width = 50
        self.assertEqual(pic.width, 50)

    def test_picture_object(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(pic.name, self.wb1.sheets[0].pictures["pic1"].name)

    def test_height(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(int(pic.height), 30)
        pic.height = 50
        self.assertEqual(int(pic.height), 50)

    def test_delete(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertTrue("pic1" in self.wb1.sheets[0].pictures)
        pic.delete()
        self.assertFalse("pic1" in self.wb1.sheets[0].pictures)

    def test_duplicate(self):
        with self.assertRaises(xw.ShapeAlreadyExists):
            filename = os.path.join(this_dir, "sample_picture.png")
            pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic1")
            pic2 = self.wb1.sheets[0].pictures.add(filename, name="pic1")

    def test_picture_update(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        pic1.update(filename)

    @unittest.skipIf(pathlib is None, "pathlib unavailable")
    def test_picture_update_pathlib(self):
        filename = pathlib.Path(this_dir) / "sample_picture.png"
        pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        pic1.update(filename)

    def test_picture_auto_update(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic1", update=True)
        pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic1", update=True)
        self.assertEqual(len(self.wb1.sheets[0].pictures), 1)

    def test_picture_auto_update_without_name(self):
        with self.assertRaises(ValueError):
            filename = os.path.join(this_dir, "sample_picture.png")
            pic1 = self.wb1.sheets[0].pictures.add(filename, update=True)

    def test_picture_index(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        self.assertEqual(
            self.wb1.sheets[0].pictures[0], self.wb1.sheets[0].pictures["pic1"]
        )
        self.assertEqual(self.wb1.sheets[0].pictures(1), self.wb1.sheets[0].pictures[0])

    def test_len(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic1")
        pic2 = self.wb1.sheets[0].pictures.add(filename, name="pic2")
        self.assertEqual(len(self.wb1.sheets[0].pictures), 2)

    def test_iter(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        names = ["pic1", "pic2"]
        pic1 = self.wb1.sheets[0].pictures.add(filename, name=names[0])
        pic2 = self.wb1.sheets[0].pictures.add(filename, name=names[1])
        for ix, pic in enumerate(self.wb1.sheets[0].pictures):
            self.assertEqual(self.wb1.sheets[0].pictures[ix].name, names[ix])

    def test_contains(self):
        filename = os.path.join(this_dir, "sample_picture.png")
        pic1 = self.wb1.sheets[0].pictures.add(filename, name="pic 1")
        self.assertTrue("pic 1" in self.wb1.sheets[0].pictures)


@unittest.skipIf(mpl is None, "matplotlib missing")
class TestMatplotlib(TestBase):
    def test_add_no_name(self):
        fig = plt.figure()
        plt.plot([-1, 1, -2, 2, -3, 3, 2])
        self.wb1.sheets[0].pictures.add(fig)
        self.assertEqual(len(self.wb1.sheets[0].pictures), 1)

    def test_add_with_name(self):
        fig = plt.figure()
        plt.plot([-1, 1, -2, 2, -3, 3, 2])
        self.wb1.sheets[0].pictures.add(fig, name="Test1")
        self.assertEqual(self.wb1.sheets[0].pictures[0].name, "Test1")


@unittest.skipIf(plotly_go is None, "plotly missing")
class TestPlotly(TestBase):
    def get_plotly_fig(self):
        N = 100
        x = np.random.rand(N)
        y = np.random.rand(N)
        colors = np.random.rand(N)
        sz = np.random.rand(N) * 30

        fig = plotly_go.Figure()
        fig.add_trace(
            plotly_go.Scatter(
                x=x,
                y=y,
                mode="markers",
                marker=plotly_go.scatter.Marker(
                    size=sz, color=colors, opacity=0.6, colorscale="Viridis"
                ),
            )
        )
        return fig

    def test_add_no_name(self):
        self.wb1.sheets[0].pictures.add(self.get_plotly_fig())
        self.assertEqual(len(self.wb1.sheets[0].pictures), 1)

    def test_add_with_name(self):
        self.wb1.sheets[0].pictures.add(self.get_plotly_fig(), name="Test1")
        self.assertEqual(self.wb1.sheets[0].pictures[0].name, "Test1")


class TestCharts(TestBase):
    def test_add_properties(self):
        sht = self.wb1.sheets[0]
        sht.range("A1").value = [["one", "two"], [1.1, 2.2]]

        self.assertEqual(len(sht.charts), 0)
        chart = sht.charts.add()
        self.assertEqual(len(sht.charts), 1)

        chart.name = "My Chart"
        chart.set_source_data(sht.range("A1").expand("table"))
        chart.chart_type = "line"

        self.assertEqual("My Chart", chart.name)
        self.assertEqual(sht.charts[0].chart_type, "line")

        chart.chart_type = "pie"
        self.assertEqual(sht.charts[0].chart_type, "pie")

        for a in ("left", "top", "width", "height"):
            setattr(chart, a, 400)
            self.assertEqual(getattr(sht.charts[0], a), 400)
            setattr(sht.charts[0], a, 500)
            self.assertEqual(getattr(chart, a), 500)

        chart.delete()
        self.assertEqual(sht.charts.count, 0)


class TestChart(TestBase):
    def test_len(self):
        chart = self.wb1.sheets[0].charts.add()
        self.assertEqual(len(self.wb1.sheets[0].charts), 1)

    def test_count(self):
        chart = self.wb1.sheets[0].charts.add()
        self.assertEqual(
            len(self.wb1.sheets[0].charts), self.wb1.sheets[0].charts.count
        )


if __name__ == "__main__":
    unittest.main()



# Allows you to use xlwings.reports instead of xlwings.pro.reports

from .pro.reports import *  # noqa: F401,F403

__all__ = (  # noqa: F405
    "render_template",
    "create_report",
    "Image",
    "Markdown",
    "MarkdownStyle",
    "formatter",
)



import argparse
import os
import shutil
import subprocess
from pathlib import Path
from shlex import split
from textwrap import dedent

this_dir = Path(__file__).resolve().parent
os.chdir(this_dir)

target_dir = "dist"
webpack_config = "webpack.config.js"
source = "src/xlwings.ts"


def webpack():
    return subprocess.run(
        split("npx webpack"),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
    )


def prepend_license(path):
    content = Path(path).read_text()
    content = (
        dedent(
            """\
            /**
            * Copyright (C) 2014 - present, Zoomer Analytics GmbH. All rights reserved.
            * Licensed under BSD-3-Clause license, see: https://docs.xlwings.org/en/stable/license.html
            *
            * This file also contains code from core-js
            * Copyright (C) 2014-2023 Denis Pushkarev, Licensed under MIT license, see https://raw.githubusercontent.com/zloirock/core-js/master/LICENSE
            * This file also contains code from Webpack
            * Copyright (C) JS Foundation and other contributors, Licensed under MIT license, see https://raw.githubusercontent.com/webpack/webpack/main/LICENSE
            */
            """
        )
        + content
    )
    Path(path).write_text(content)


def build(version):
    # Clear
    if Path(target_dir).exists():
        shutil.rmtree(target_dir)

    # Version
    content = Path(source).read_text()
    content = content.replace('version = "dev"', f'version = "{version}"')
    Path(source).write_text(content)

    # Minified build
    print(webpack())
    prepend_license(f"{target_dir}/xlwings.min.js")

    # Non-minified build
    content = Path(webpack_config).read_text()
    content = content.replace('mode: "production"', 'mode: "development"').replace(
        'filename: "xlwings.min.js"', 'filename: "xlwings.js"'
    )
    Path(webpack_config).write_text(content)
    print(webpack())
    prepend_license(f"{target_dir}/xlwings.js")

    # Reset version
    content = Path(source).read_text()
    content = content.replace(f'version = "{version}"', 'version = "dev"')
    Path(source).write_text(content)

    # Reset webpack.config.js
    content = Path(webpack_config).read_text()
    content = content.replace('mode: "development"', 'mode: "production"').replace(
        'filename: "xlwings.js"', 'filename: "xlwings.min.js"'
    )
    Path(webpack_config).write_text(content)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", required=True)
    args = parser.parse_args()
    build(args.version)



from fastapi import Body, FastAPI
from fastapi.middleware.cors import CORSMiddleware

import xlwings as xw

app = FastAPI()


@app.post("/hello")
def hello(data: dict = Body):
    # Instantiate a Book object with the deserialized request body
    book = xw.Book(json=data)

    # Use xlwings as usual
    sheet = book.sheets[0]
    cell = sheet["A1"]
    if cell.value == "Hello xlwings!":
        cell.value = "Bye xlwings!"
    else:
        cell.value = "Hello xlwings!"

    # Pass the following back as the response
    return book.json()


# Excel via OfficeScripts requires CORS
app.add_middleware(
    CORSMiddleware,
    allow_origin_regex=r"https://.*.officescripts.microsoftusercontent.com",
    allow_methods=["POST"],
    allow_headers=["*"],
)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("main:app", host="127.0.0.1", port=8000, reload=True)



import unittest
from pathlib import Path

import pandas as pd

import xlwings as xw

this_dir = Path(__file__).parent


class TestTable(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.book = xw.Book()
        cls.sheet = cls.book.sheets[0]
        cls.sheet["A1"].value = [["a", "b"], [1, 2]]
        cls.test_table = cls.sheet.tables.add(source=cls.sheet["A1"].expand())

    @classmethod
    def tearDownClass(cls):
        cls.book.close()

    def test_add_table_no_name(self):
        self.assertEqual(self.sheet.tables[0].name, "Table1")

    def test_add_table_with_name(self):
        self.sheet["A4"].value = [["a", "b"], [1, 2]]
        self.sheet.tables.add(source=self.sheet["A4"].expand(), name="AABBCC")
        self.assertEqual(self.sheet.tables["AABBCC"].name, "AABBCC")

    def test_data_body_range(self):
        self.assertEqual(self.test_table.data_body_range, self.sheet["A2:B2"])

    def test_display_name(self):
        origin_display_name = self.test_table.display_name
        self.test_table.display_name = "ABCDE"
        self.assertEqual(self.test_table.display_name, "ABCDE")
        self.test_table.display_name = origin_display_name

    def test_header_row_range(self):
        self.assertEqual(self.test_table.header_row_range, self.sheet["A1:B1"])
        self.test_table.show_headers = False
        self.assertIsNone(self.test_table.header_row_range)
        self.test_table.show_headers = True

    def test_insert_row_range(self):
        table = self.sheet.tables.add(self.sheet["A10"])
        self.assertEqual(table.insert_row_range, self.sheet["A11"])

    def test_insert_row_range_none(self):
        self.assertIsNone(self.test_table.insert_row_range)

    def test_name(self):
        original_name = self.test_table.name
        self.test_table.name = "XYZ"
        self.assertEqual(self.test_table.name, "XYZ")
        self.assertEqual(self.sheet.tables["XYZ"].name, "XYZ")
        self.test_table.name = original_name

    def test_parent(self):
        self.assertEqual(self.test_table.parent, self.sheet)

    def test_show_autofilter(self):
        self.assertTrue(self.test_table.show_autofilter)
        self.test_table.show_autofilter = False
        self.assertFalse(self.test_table.show_autofilter)
        self.test_table.show_autofilter = True

    def test_show_headers(self):
        self.assertTrue(self.test_table.show_headers)
        self.test_table.show_headers = False
        self.assertFalse(self.test_table.show_headers)
        self.test_table.show_headers = True

    def test_show_table_style_columns_stripes(self):
        self.assertFalse(self.test_table.show_table_style_column_stripes)
        self.test_table.show_table_style_column_stripes = True
        self.assertTrue(self.test_table.show_table_style_column_stripes)
        self.test_table.show_table_style_column_stripes = False

    def test_show_table_style_first_column(self):
        self.assertFalse(self.test_table.show_table_style_first_column)
        self.test_table.show_table_style_first_column = True
        self.assertTrue(self.test_table.show_table_style_first_column)
        self.test_table.show_table_style_first_column = False

    def test_show_table_style_last_column(self):
        self.assertFalse(self.test_table.show_table_style_last_column)
        self.test_table.show_table_style_last_column = True
        self.assertTrue(self.test_table.show_table_style_last_column)
        self.test_table.show_table_style_last_column = False

    def test_show_table_style_row_stripes(self):
        self.assertTrue(self.test_table.show_table_style_row_stripes)
        self.test_table.show_table_style_row_stripes = False
        self.assertFalse(self.test_table.show_table_style_row_stripes)
        self.test_table.show_table_style_row_stripes = True

    def test_show_totals(self):
        self.assertFalse(self.test_table.show_totals)
        self.test_table.show_totals = True
        self.assertTrue(self.test_table.show_totals)
        self.test_table.show_totals = False

    def test_table_style(self):
        self.assertEqual(self.test_table.table_style, "TableStyleMedium2")
        self.test_table.table_style = "TableStyleMedium1"
        self.assertEqual(self.test_table.table_style, "TableStyleMedium1")
        self.test_table.table_style = "TableStyleMedium2"

    def test_totals_row_range(self):
        self.assertIsNone(self.test_table.totals_row_range)
        self.test_table.show_totals = True
        self.assertEqual(self.test_table.totals_row_range, self.sheet["A3:B3"])
        self.test_table.show_totals = False

    def test_resize(self):
        self.assertEqual(self.test_table.range.address, "$A$1:$B$2")
        self.test_table.resize(self.sheet["A1:C3"])
        self.assertEqual(self.test_table.range.address, "$A$1:$C$3")
        self.test_table.resize(self.sheet["$A$1:$B$2"])
        self.assertEqual(self.test_table.range.address, "$A$1:$B$2")


class TestTableUpdate(unittest.TestCase):
    def test_table_update(self):
        df = pd.DataFrame(
            {
                "a": [1, 2, 3, 4, 5],
                "b": [11, 22, 33, 44, 55],
                "c": [111, 222, 333, 444, 555],
                "d": [1111, 2222, 3333, 4444, 5555],
            }
        )
        book = xw.Book(this_dir / "tables.xlsx")
        sheet = book.sheets["template"].copy()
        sheet.tables[0].update(df)
        sheet.tables[1].update(df)
        sheet.tables[2].update(df)
        sheet.tables[3].update(df, index=False)
        self.assertEqual(sheet["A1:E50"].value, book.sheets["expected"]["A1:E50"].value)
        sheet.book.close()


if __name__ == "__main__":
    unittest.main()



"""
This is run by GH Actions to make sure the version corresponds to the release version
"""

import os
from pathlib import Path

this_dir = Path(__file__).resolve().parent


def main():
    if os.environ["GITHUB_REF"].startswith("refs/tags"):
        version = os.environ["GITHUB_REF"][10:]
    else:
        return

    xlwingsjs = this_dir / "dist" / "xlwings.js"
    xlwingsminjs = this_dir / "dist" / "xlwings.min.js"

    if f'version = "{version}"' not in xlwingsjs.read_text():
        raise Exception("Didn't find expected version in xlwings.js!")

    if f'version="{version}"' not in xlwingsminjs.read_text():
        raise Exception("Didn't find expected version in xlwings.min.js!")


if __name__ == "__main__":
    main()



import os
import sys
import time
import unittest

import xlwings as xw

from .common import SPEC, TestBase, this_dir


class TestApps(TestBase):
    def test_active(self):
        self.assertTrue(xw.apps.active in [self.app1, self.app2])

    def test_len(self):
        n_original = len(xw.apps)
        app = xw.App(spec=SPEC)
        wb = app.books.add()
        self.assertEqual(n_original + 1, len(xw.apps))
        app.quit()

    def test_count(self):
        self.assertEqual(xw.apps.count, len(xw.apps))

    def test_iter(self):
        for app in xw.apps:
            if app == (self.app1 or self.app2):
                self.assertEqual(len(app.books), 2)

    def test_keys(self):
        k = xw.apps.keys()[0]
        self.assertEqual(xw.apps[k], xw.apps(k))


class TestApp(TestBase):
    def test_activate(self):
        if sys.platform.startswith("win") and self.app1.version.major > 14:
            # Excel >= 2013 on Win has issues with activating hidden apps correctly
            # over two instances
            with self.assertRaises(Exception):
                self.app1.activate()
        else:
            self.assertEqual(self.app2, xw.apps.active)
            self.app1.activate()
            self.assertEqual(self.app1, xw.apps.active)

    def test_visible(self):
        # Can't successfully test for False on Mac...?
        self.app1.visible = True
        self.assertTrue(self.app1.visible)

    def test_quit(self):
        app = xw.App()
        n_apps = len(xw.apps)
        app.quit()
        time.sleep(1)  # needed for Mac Excel 2011
        self.assertEqual(n_apps - 1, len(xw.apps))

    def test_kill(self):
        app = xw.App(spec=SPEC)
        n_apps = len(xw.apps)
        app.kill()
        import time

        time.sleep(0.5)
        self.assertEqual(n_apps - 1, len(xw.apps))

    def test_screen_updating(self):
        self.app1.screen_updating = False
        self.assertEqual(self.app1.screen_updating, False)

        self.app1.screen_updating = True
        self.assertTrue(self.app1.screen_updating)

    def test_display_alerts(self):
        self.app1.display_alerts = False
        self.assertEqual(self.app1.display_alerts, False)

        self.app1.display_alerts = True
        self.assertTrue(self.app1.display_alerts)

    def test_enable_events(self):
        self.app1.enable_events = False
        self.assertEqual(self.app1.enable_events, False)

        self.app1.enable_events = True
        self.assertTrue(self.app1.enable_events)

    # TODO: refactor test to catch exception on macOS
    @unittest.skipIf(
        sys.platform.startswith("darwin"), "app.interactive is not supported on macOS"
    )
    def test_interactive(self):
        self.app1.interactive = False
        self.assertEqual(self.app1.interactive, False)

        self.app1.interactive = True
        self.assertTrue(self.app1.interactive)

    def test_calculation_calculate(self):
        sht = self.wb1.sheets[0]
        sht.range("A1").value = 2
        sht.range("B1").formula = "=A1 * 2"

        self.app1.calculation = "manual"
        sht.range("A1").value = 4
        self.assertEqual(sht.range("B1").value, 4)

        self.app1.calculation = "automatic"
        # This is needed on Mac Excel 2016 but not onMac Excel 2011 (changed behaviour)
        self.app1.calculate()
        self.assertEqual(sht.range("B1").value, 8)

        sht.range("A1").value = 2
        self.assertEqual(sht.range("B1").value, 4)

    def test_calculation(self):
        self.app1.calculation = "automatic"
        self.assertEqual(self.app1.calculation, "automatic")

        self.app1.calculation = "manual"
        self.assertEqual(self.app1.calculation, "manual")

        self.app1.calculation = "semiautomatic"
        self.assertEqual(self.app1.calculation, "semiautomatic")

    def test_version(self):
        self.assertTrue(self.app1.version.major > 0)

    def test_wb_across_instances(self):
        app1_wb_count = len(self.app1.books)
        app2_wb_count = len(self.app2.books)

        wb2 = self.app1.books.add()
        wb3 = self.app2.books.add()
        wb4 = self.app2.books.add()
        wb5 = self.app2.books.add()

        self.assertEqual(len(self.app1.books), app1_wb_count + 1)
        self.assertEqual(len(self.app2.books), app2_wb_count + 3)

        wb2.close()
        wb3.close()
        wb4.close()
        wb5.close()

    def test_selection(self):
        self.assertEqual(self.app1.selection.address, "$A$1")

    def test_books(self):
        self.assertEqual(len(self.app2.books), 2)

    def test_pid(self):
        self.assertTrue(self.app1.pid > 0)

    def test_len(self):
        n_books = len(self.app1.books)
        self.app1.books.add()
        self.assertEqual(len(self.app1.books), n_books + 1)

    def test_macro(self):
        wb = self.app1.books.open(os.path.join(this_dir, "macro book.xlsm"))
        test1 = self.app1.macro("Module1.Test1")
        res1 = test1("Test1a", "Test1b")
        self.assertEqual(res1, 1)


class TestAppPropertiesContextManager(unittest.TestCase):
    def test_properties_context_manager(self):
        book = xw.Book()
        app = book.app
        self.assertTrue(app.display_alerts)
        self.assertTrue(app.enable_events)
        with app.properties(display_alerts=False):
            self.assertFalse(app.display_alerts)
            self.assertTrue(app.enable_events)
            with app.properties(display_alerts=True, enable_events=False):
                self.assertTrue(app.display_alerts)
                self.assertFalse(app.enable_events)
            self.assertFalse(app.display_alerts)
            self.assertTrue(app.enable_events)
        self.assertTrue(app.display_alerts)
        self.assertTrue(app.enable_events)
        book.close()


if __name__ == "__main__":
    unittest.main()



"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

from .embedded_code import dump_embedded_code, runpython_embedded_code
from .module_permissions import verify_execute_permission
from .reports import Markdown, MarkdownStyle
from .udfs_officejs import (
    custom_functions_call,
    custom_functions_code,
    custom_functions_meta,
    xlarg as arg,
    xlfunc as func,
    xlret as ret,
)
from .utils import LicenseHandler

__all__ = (
    "dump_embedded_code",
    "runpython_embedded_code",
    "verify_execute_permission",
    "Markdown",
    "MarkdownStyle",
    "arg",
    "func",
    "ret",
    "custom_functions_code",
    "custom_functions_meta",
    "custom_functions_call",
)

LicenseHandler.validate_license("pro")



"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

"""
This engine is only used in connection with Office.js UDFs, not with runPython.
"""

import datetime as dt

try:
    import numpy as np
except ImportError:
    np = None
try:
    import pandas as pd
except ImportError:
    pd = None

from .. import utils


def datetime_to_formatted_number(datetime_object, date_format, runtime):
    # https://learn.microsoft.com/en-us/javascript/api/requirement-sets/excel/custom-functions-requirement-sets
    serial = utils.datetime_to_xlserial(datetime_object)
    if float(runtime) < 1.4:
        return serial
    else:
        return {
            "type": "FormattedNumber",
            "basicValue": serial,
            "numberFormat": date_format,
        }


def errorstr_to_errortype(error, runtime):
    if float(runtime) < 1.4:
        return error
    error_to_type = {
        "#DIV/0!": "Div0",
        "#N/A": "NotAvailable",
        "#NAME?": "Name",
        "#NULL!": "Null",
        "#NUM!": "Num",
        "#REF!": "Ref",
        "#VALUE!": "Value",
    }
    error_type = error_to_type.get(error)
    if not error_type:
        return error
    else:
        return {
            "type": "Error",
            "errorType": error_type,
        }


def _clean_value_data_element(
    value,
    datetime_builder,
    empty_as,
    number_builder,
    err_to_str,
):
    # datetime_builder is not supported as normal date-formatted cells aren't recognized
    if value == "":
        return empty_as
    elif isinstance(value, dict):
        # https://learn.microsoft.com/en-us/office/dev/add-ins/excel/custom-functions-data-types-concepts
        if value["type"] == "Error":
            if err_to_str:
                return value["basicValue"]
            else:
                return None
        else:
            value = value["basicValue"]  # e.g., datetime (only via data types)
    elif number_builder is not None and type(value) == float:
        value = number_builder(value)
    return value


class Engine:
    @staticmethod
    def clean_value_data(data, datetime_builder, empty_as, number_builder, err_to_str):
        return [
            [
                _clean_value_data_element(
                    c, datetime_builder, empty_as, number_builder, err_to_str
                )
                for c in row
            ]
            for row in data
        ]

    @staticmethod
    def prepare_xl_data_element(x, options):
        if x is None:
            return ""
        elif pd and pd.isna(x):
            return errorstr_to_errortype("#NUM!", options["runtime"])
        elif np and isinstance(x, (np.floating, float)) and np.isnan(x):
            return errorstr_to_errortype("#NUM!", options["runtime"])
        elif np and isinstance(x, np.number):
            return float(x)
        elif np and isinstance(x, np.datetime64):
            return datetime_to_formatted_number(
                utils.np_datetime_to_datetime(x),
                options["date_format"],
                options["runtime"],
            )
        elif pd and isinstance(x, pd.Timestamp):
            return datetime_to_formatted_number(
                x.to_pydatetime(),
                options["date_format"],
                options["runtime"],
            )
        elif pd and isinstance(x, type(pd.NaT)):
            # This seems to be caught by pd.isna() nowadays?
            return ""
        elif isinstance(x, (dt.date, dt.datetime)):
            return datetime_to_formatted_number(
                x, options["date_format"], options["runtime"]
            )
        elif isinstance(x, str) and x.startswith("#"):
            return errorstr_to_errortype(x, options["runtime"])
        return x

    @property
    def name(self):
        return "officejs"

    @property
    def type(self):
        return "remote"


engine = Engine()
